@@ -12,7 +12,7 @@ package Net::SIP::StatelessProxy;
 use fields qw( dispatcher rewrite_contact nathelper force_rewrite );
 
 use Net::SIP::Util ':all';
-use Digest::MD5 qw(md5_hex md5);
+use Digest::MD5 qw(md5);
 use Carp 'croak';
 use List::Util 'first';
 use Hash::Util 'lock_ref_keys';
@@ -29,6 +29,10 @@ use Net::SIP::Debug;
 #        if called on a string without @ which cannot rewritten back it
 #        should return undef. If not given a reasonable default will be
 #        used.
+#     rewrite_crypt: function(data,dir,add2mac) which will encrypt(dir>0) or
+#        decrypt(dir<0) data. Optional add2mac is added in MAC. Will return
+#        encrypted/decrypted data or undef if decryption failed because
+#        MAC did not match
 #     nathelper: Net::SIP::NAT::Helper used for rewrite SDP bodies.. (optional)
 #     force_rewrite: if true rewrite contact even if incoming and outgoing
 #         legs are the same
@@ -54,46 +58,43 @@ sub new {
 # default handler for rewriting, does simple XOR only,
 # this is not enough if you need to hide internal addresses
 sub _default_rewrite_contact {
-    my ($crypt,$disp,$contact,$leg_in,$leg_out) = @_;
+    my ($crypt,$disp,$contact,$leg_in,$leg_out,$force_rewrite) = @_;
 
-    if ( $contact =~m{\@} ) {
+    my $legdict;
+    my ($ileg_in,$ileg_out) = $disp->legs2i($leg_in,$leg_out,\$legdict);
+
+    if ($force_rewrite or $contact =~m{\@}) {
 	# needs to be rewritten - incorporate leg_in:leg_out
-	$contact = join("\|",
-	    (map { $_->key } ($leg_in,$leg_out)),
-	    $contact
-	);
-	# add 'r' in front of hex so it does not look like phone number
-	my $new = 'r'.unpack( 'H*',$crypt->($contact,1));
+	$contact = pack("nna*",$ileg_in,$ileg_out,$contact);
+	# add 'b' in front so it does not look like phone number
+	my $new = 'b'._encode_base32($crypt->($contact,1,$legdict));
 	DEBUG( 100,"rewrite $contact -> $new" );
 	return $new;
     }
 
-    if ( $contact =~m{^r([0-9a-f]+)$} ) {
+    if ( $contact =~m{^b([A-Z2-7]+)$} ) {
 	# needs to be written back
-	my $old = $crypt->(pack("H*",$1),-1) or do {
+	my $old = $crypt->(_decode_base32($1),-1,$legdict) or do {
 	    DEBUG(10,"no rewriting of $contact - bad encryption");
 	    return;
 	};
 	DEBUG(100,"rewrote back $contact -> $old");
-	(my $old_in,my $old_out,$old) = split( m{\|},$old,3);
+	(my $iold_in,my $iold_out,$old) = unpack("nna*",$old);
 	my $new_in = $leg_in->key;
-	if ( $new_in ne $old_out ) {
-	    DEBUG(10,"no rewriting of $contact - went out through $old_out, came in through $new_in");
+	if ($ileg_in ne $iold_out) {
+	    DEBUG(10,"no rewriting of $contact - went out through $iold_out, came in through $ileg_in");
 	    return;
 	}
 	if ( ref($leg_out) eq 'SCALAR' ) {
 	    # return the old_in as the new outgoing leg
-	    my @l = grep { $_->key eq $old_in } $disp->get_legs;
-	    if ( @l != 1 ) {
-		DEBUG(10,"no rewriting of $contact - cannot find leg $old_in");
+	    ($$leg_out) = $disp->i2legs($iold_in) or do {
+		DEBUG(10,"no rewriting of $contact - cannot find leg $iold_in");
 		return;
 	    }
-	    $$leg_out = $l[0];
-	} elsif ( $leg_out ) {
+	} elsif ($leg_out) {
 	    # check that it is the expected leg
-	    my $new_out = $leg_out->key;
-	    if ( $new_out ne $old_in ) {
-		DEBUG(10,"no rewriting of $contact - went in through $old_in, should got out through $new_out");
+	    if ($ileg_out ne $iold_in) {
+		DEBUG(10,"no rewriting of $contact - went in through $iold_in, should got out through $ileg_out");
 		return;
 	    }
 	}
@@ -107,19 +108,35 @@ sub _default_rewrite_contact {
 }
 
 {
+    # This is only a simple implementation which is in no way cryptographic safe
+    # because it does use a broken cipher (RC4), pseudo-random keys and IV only
+    # and short keys. Nonetheless, it is probably safe for this purpose and does
+    # not depend on non-standard libs, but using openssl bindings might be both
+    # more secure and faster for this.
+    #
     # RC4 with seed + checksum, picks random key on first use
     # dir: encrypt(1),decrypt(-1), otherwise symmetric w/o seed and checksum
-    my @k;
+    my (@k,$mackey);
     sub _stupid_crypt {
-	my ($in,$dir) = @_;
-	@k = map { rand(256) } (0..20) if ! @k; # create random key 
+	my ($in,$dir,$add2mac) = @_;
+	$add2mac = '' if ! defined $add2mac;
+
+	if (!@k) {
+	    # create random key
+	    @k = map { rand(256) } (0..20);
+	    $mackey = pack("N",rand(2**32));
+	}
 
 	if ($dir>0) {
 	    $in = pack("N",rand(2**32)).$in;  # add seed
-	    $in .= substr(md5($in),0,4);      # add checksum
+	} else {
+	    # remove checksum and verify it
+	    my $cksum = substr($in,-4,4,'');
+	    substr(md5($in.$add2mac.$mackey),0,4) eq $cksum
+		or return;  # does not match
 	}
 
-	# RC4
+	# apply RC4 for encryption/decryption
 	my $out = '';
 	my @s = (0..255);
 	my $x = my $y = 0;
@@ -135,13 +152,35 @@ sub _default_rewrite_contact {
 	    $out .= pack('C',$_^=$s[($s[$x]+$s[$y])%256]);
 	}
 
-	if ( $dir<0 ) {
-	    my $cksum = substr($out,-4,4,'');           # remove checksum
-	    substr(md5($out),0,4) eq $cksum or return;  # verify it
-	    substr($out,0,4,'');                        # remove seed
+	if ($dir>0) {
+	    # add checksum
+	    $out .= substr(md5($out.$add2mac.$mackey),0,4);
+	} else {
+	    substr($out,0,4,'');  # remove seed
 	}
 	return $out;
     }
+
+    sub _encode_base32 {
+	my $data = shift;
+	$data = unpack('B*',$data);
+	my $text;
+	my $padsize =
+	$data .= '0' x ((5 - length($data) % 5) % 5); # padding
+	$data =~s{(.....)}{000$1}g;
+	$data = pack('B*',$data);
+	$data =~tr{\000-\037}{A-Z2-7};
+	return $data;
+    }
+
+    sub _decode_base32 {
+	my $data = shift;
+	$data =~ tr{A-Z2-7a-z}{\000-\037\000-\031};
+	$data = unpack('B*',$data);
+	$data =~s{...(.....)}{$1}g;
+	$data = substr($data,0,8*int(length($data)/8));
+	return pack('B*',$data);
+    }
 }
 
 ###########################################################################
@@ -516,7 +555,7 @@ sub __forward_packet_final {
 	    # otherwise rewrite it
 	    } else {
 		$addr = invoke_callback($rewrite_contact,$addr,$incoming_leg,
-		    $outgoing_leg);
+		    $outgoing_leg,1);
 		$addr .= '@'.$outgoing_leg->laddr(2);
 		my $cnew = sip_parts2hdrval( 'contact', $pre.$addr.$post, $p );
 		DEBUG( 50,"rewrote '$c' to '$cnew'" );
